/*
* user.cpp - interface with API.
*
* Copyright 1999-2000 by Columbia University; all rights reserved 
*/

#include <time.h>      /* time() */
#include <sys/time.h>  /* gettimeofday */
#include <errno.h>     /* ETIME */
#include <stdlib.h>    /* calloc() */
#include <string.h>    /* strerror(), strcmp() */

extern "C" {
#include <unistd.h>    /* gettimeofday() */
#include <pthread.h>   /* pthread_create, ... */
#include "config.h"
#include "sysdep.h"
#include "policy.h"
#include "strdupn.h"
#include "error.h"
#include "proxy-support.h"
}

#include "sipcall.h"
#include "sipreg.h"
#include "sipintf.h"

/* Local prototype */
static void NextUserProxyStep(request_t *r, sipcall_t *call);

/* Extern prototypes. */
extern "C" unsigned long random32();

/* 
 * User Policy: handle "generic" API requests/responses 
 */
struct user_info {
  enum { sending_requests, awaiting_responses, final_wait} state;
  int pos;
  int more;
  policy_info_t *reg;
  char *callid;
};

/*
 * StartSomeRegBranches()
 * Start some branches based on the registration structure -- those
 * that are within "spread" of the first non-started branch.
 * Returns whether there are still more branches to start after this.
 * XXX this is same as proxy.c.
 */
static int StartSomeRegBranches(request_t *r, const policy_info_t *reg,
                                int *pos)
{
  double q;              /* current quality value */
  double spread = 0.05;  /* spread of things to do in parallel */
  unsigned int i;
  message_t *m;

  q = reg->contact[*pos].q;

  for (i = *pos; i < reg->contacts && q - reg->contact[i].q <= spread; i++) {
    m = GetRequestAsMessage(r);
    policy_proxy_original_request_to(r, &reg->contact[i].uri, m, NULL,
                                     transaction_policy);
    MessageDelete(m);
  }
  *pos = i;
  return (i < reg->contacts);
} /* StartSomeRegBranches */


extern "C" int RFC822_ParseLine(unsigned char *line, int linelength,
  parser_header_t **php);

/* 
 * userapi_init()
 * Called when a transaction being handled by user_policy is started
 * This assumes that we're only called if there's at least one address
 * which it's reasonable to proxy to.  XXX: check what happens if we
 * register with mailto addresses, etc. This creates a call thread for 
 * the call if needed.
 */
static transaction_policy_ret userapi_init(request_t *r,
                                           const policy_info_t *reg,
                                           int status)
{
  struct user_info *info;

  info = (struct user_info*) calloc(1, sizeof(struct user_info));
  if (info == NULL) {
    error("userapi_init", "calloc error\n");
    return transaction_policy_error;
  }
  info->callid = SIP_CallID(&r->p);
  if (info->callid != NULL) {
    info->callid = strdup(info->callid);
  }

  r->transaction_policy_info = (void *)info;
  
  if (reg != NULL) {
    /* 
     * This is an outgoing request.
     * Set up the request structure so branches can be invoked.
     */
    /* Compute the request hash */
    r->hash         = ComputeRequestHash(r);

    /* Allocate and copy the registration struct.
     * This is because the called may pass a pointer to the
     * local variable which may not be valid in the threads
     * when the original calling function is terminated.
     * do this in proxy.c also XXX 
     */
    info->reg   = (policy_info_t *) calloc(1, sizeof(policy_info_t));
    memcpy(info->reg,reg,sizeof(policy_info_t));

    info->state = user_info::sending_requests;
  }
  else {
    info->reg = NULL;
  }

  return transaction_policy_ok;
} /* userapi_init */


/*
 * userapi_handle_initial_request()
 * Called when a transaction being handled by user_policy is started
 * This assumes that we're only called if there's at least one address
 * which it's reasonable to proxy to.  XXX: check what happens if we
 * register with mailto addresses, etc. This creates a call thread for 
 * the call if needed.
 */
static transaction_policy_ret userapi_handle_initial_request(request_t *r)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;
  sipcall_t *call;

  if (info->reg == NULL) {
    int tagadded = 0;

    /*
     * This is an incoming request
     * call has already been created.
     * Just signal the incoming request to
     * the call thread.
     */
    call = CallSearch(info->callid);
    if (call == NULL) {
      error("userapi_init", "No call present\n");
      return transaction_policy_error;
    }

    /*
     * Add the tag in the To header if it is not already there.
     * Should we check if Number of via headers are more than 1.
     * I don't think we need to.
     */
    parser_header_t *h;  /* request header */
    if ((h = RFC822_Header(r->p.headers, "To")) && h->elements) {
      /* Update the tag */
      char *tag = RFC822_Attribute1(h->elements->parameters, "tag");
      if (!tag || !(*tag)) {
        RFC822_AddAttribute(h, "tag", SIPLibrary::UAID, PT_token);
        /* Store it internally for subsequent request responses */
        strassign(&call->self_tag, SIPLibrary::UAID);
        tagadded = 1;
      }
      else {
        if (call->self_tag == NULL || !strcmp(tag, call->self_tag)) {
          /* Matching tag, good ! */
          tagadded = 1;
        }
        else {
          /* Tag does not match, fail with 481 */
          error("userapi_init", "Tag mismatch %s != %s\n",
                         call->self_tag , tag);
          MessageEnqueuePseudoResponse(&r->mqueue, NULL, nobody_handled, 481);
          pthread_cond_broadcast(&r->cond);
          tagadded = 0;
        }
      }
    }
    if ((h= RFC822_Header(r->p.headers, "From")) && h->elements) {
      /* Update the remote tag */
      char *tag = RFC822_Attribute1(h->elements->parameters, "tag");
      if (tag && *tag) {
        if (call->remote_tag != NULL &&
             strcmp(call->remote_tag, tag) != 0) {
          /* Tag mismatch */
          error("userapi_init", "Tag mismatch %s != %s\n",
                          call->remote_tag, tag);
          MessageEnqueuePseudoResponse(&r->mqueue, NULL, nobody_handled, 481);
          pthread_cond_broadcast(&r->cond);
          tagadded = 0;
        }
        else if (call->remote_tag == NULL) {
          strassign(&call->remote_tag, tag);
          /* Do not set tagadded, since that will undo the To tag
           * matching */
        }
        else {
          /* Matching tag, fall through */
          /* Do not set tagadded, since that will undo the To tag
           * matching */
        }
      }
    }

    if (tagadded) {
      /* Format the message and send */
      //message = GetRequestAsMessage(r);

      /* Hand it up to the call thread. */
      MessageEnqueue(&call->mqueue, message_real_message, &r->p,
                    &r->request_body, &r->source_sin, 
                     NULL,
                     nobody_handled);
      pthread_cond_broadcast(&call->cond);
      pthread_mutex_unlock(&call->mutex);
    }
    else {
      /* Do do do anything, the policy thread
       * will wait for ACK */
    }

    /* Set the timeout for incoming request also */
    policy_set_timeout(r, (double) config.timeout, transaction_policy);
  }
  else {
    /* 
     * This is an outgoing request.
     * STart the branches (UAC) to send the
     * outgoing requests.
     * The code is similar to the proxy_policy code.
     */

    /* XXX: For sending out MESSAGE */
    if (strcmp(r->p.method, "MESSAGE") == 0){
      debug(DEBUG_MISC, "userapi_handle_initial_request", "set r->sent_final\n");
      r->sent_final = 1;
    }

    info->more  = StartSomeRegBranches(r, info->reg, &info->pos);
    if (!info->more) {
      info->state = user_info::awaiting_responses;
    }
    policy_set_timeout(r, (double) config.timeout, transaction_policy);
  }

  return transaction_policy_ok;
}


/*
 * userapi_handle_subsequent_request()
 * Called when a follow-up request (ACK or CANCEL) comes in from
 * the client. This is also called
 * when the CANCEL or ACK is requested by the call thread.
 * Dispatch the request as appropriate to the call thread
 * or the UAC clients appropriately.
 */
static transaction_policy_ret userapi_handle_subsequent_request(request_t *r, message_t *m)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;
  sipcall_t *call = NULL;

  debug(DEBUG_MISC, "userapi_handle_request","called\n");

  if (info->reg == NULL) {
    debug(DEBUG_MISC, "userapi_handle_request", "Incoming request\n");
    /* 
     * received new request from remote 
     * forward this request to the call thread.
     */
    call = CallSearch(info->callid);
    if (call == NULL) {
      error("userapi_handle_request", "No call present\n");
      m->handled = finished_handling;
      return transaction_policy_ok;
    }

    if (strcmp(m->p.method, "CANCEL") == 0) {
      /* 
       * We were canceled.  inform the call thread
       * and send back 487.
       */

      if (info->state != user_info::final_wait) {
        policy_send_new_response(r, 487, NULL);
        info->state = user_info::final_wait;
        policy_set_timeout(r, (double)config.timeout, transaction_policy);
      }

      /* Hand it up to the call thread. */
      MessageEnqueue(&call->mqueue, message_real_message, &m->p,
                     &m->body, &m->source, 
                     NULL,
                     nobody_handled);
      pthread_cond_broadcast(&call->cond);
      m->handled = finished_handling;
    }
    else if (strcmp(m->p.method, "ACK") == 0) {
      debug(DEBUG_MISC, "userapi_handle_request", "ACK arrived\n");
      r->acked = 1; /* The request is acked */
      if (!r->sent_200) {
        /* Receive ACK without sending 2xx */
        error("userapi_handle_request", "ACK arrived at unexpected time\n");
      }
      else {
       /* Hand it up to the call thread. */
        /* XXX should we remove content-type and 
         * set content-length to 0 here */
       MessageEnqueue(&call->mqueue, message_real_message, &m->p,
                       &m->body, &m->source, 
                       NULL,
                       nobody_handled);
       pthread_cond_broadcast(&call->cond);

      }
      m->handled = finished_handling;
    }
    else {
      /* 
       * I don't think any other methods should be showing
       * up here. 
       */
      error("userapi_handle_request",
            "Request with surprising method %s on message queue\n",
            m->p.method);
      m->handled = finished_handling;
    }

    debug(DEBUG_MISC, "userapi_handle_request", "Message sent\n");
    pthread_mutex_unlock(&call->mutex);
  }
  else {
    debug(DEBUG_MISC, "userapi_handle_request", "Outgoing request\n");

    /* This message is from call thread. */
    if (strcmp(m->p.method, "CANCEL") == 0) {
      /* 
       * We were canceled.  cancel all the unfinished
       * branches. XXX what about finished branches ?
       * This function is defined in proxy.c.
       */

      CancelUnfinishedBranches(r);

      m->handled = finished_handling;
    }
    else if (strcmp(m->p.method, "ACK") == 0) {
      r->acked = 1; /* The request is acked */
      if (!r->sent_200) {
        /* Receive ACK without receiving 2xx */
        error("userapi_handle_request", "ACK arrived at unexpected time");
      }
      else {
        if (m->type == message_pseudo_message) {
          /* This should not happen because
           * the ACk message is formatted and sent by the
           * call thread, user.cpp */
          /* build the message */
          HTTP_Duplicate(&m->p, &r->p);
          m->type = message_real_message;
          strassign(&m->p.method, "ACK"); /* set method to ACK */
          /* Change the CSeq method and
           * the Content-Type and Content-Length XXX */
        }
        else {
          /* ACK is already built */
        }

        /* Hand it up to the call thread. 
         * This method is defined in proxy.c file */
        ForwardACKToBranches(r, m);
      }
      m->handled = finished_handling;
    }
    else {
      /* 
       * I don't think any other methods should be showing
       * up here. 
       */
      error("userapi_handle_request",
            "Request with surprising method %s on message queue\n",
            m->p.method);
      m->handled = finished_handling;
    }
  }
  return transaction_policy_ok;
} /* userapi_handle_request */


/* Do3xxRecursion()
 * We have received a 3xx response.  Go through it and see if it has any
 * Contact headers with SIP URI's.  If it does, immediately start new
 * child branches to those URI's, and return 1.  If not, return 0.
 * XXX: update registry on 301, 305?
 * This is copied from proxy.c. XXX merge ?
 */
static int Do3xxRecursion(request_t *r, message_t *m)
{
  int found_something = 0, already_tried;
  sip_contact_t *contact;
  int i, j;
  message_t *req_copy;
  
  contact = SIP_Contact(&m->p);
  if (!contact)
    return 0;

  for (i = 0; contact[i].uri.scheme != NULL; i++) {
    if (strcmp(contact[i].uri.scheme, "sip") == 0) {
      found_something = 1;
      /* Check if some other branch has already visited this URI */
      already_tried = 0;
      for (j = 0; j < r->branch_count; j++) {
        if (URI_Equal_With_Defaults(r->branches[j].uri, contact[i].uri,
                                    sip_uri_defaults) == 0) {
          already_tried = 1;
          break;
        }
      }
      if (!already_tried) {
        req_copy = GetRequestAsMessage(r);
        policy_proxy_original_request_to(r, &contact[i].uri, req_copy, NULL,
                                         transaction_policy);
        MessageDelete(req_copy);
      }
    }
  }

  SIP_FreeContact(contact);
  return found_something;
}

/*
 * userapi_handle_response()
 * Called when a response comes in from some branch or
 * the call thread responds for an incoming request.
 * Dispatch the response as appropriate to the
 * remote or the call thread.
 */
static transaction_policy_ret userapi_handle_response(request_t *r, message_t *m,
                                          branch_t *b)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;
  sipcall_t *call = NULL;

  if (info->reg == NULL) {
    /*
     * This response if sent by the call thread for remote
     * for an incoming request.
     */

    if (m->p.status < 100 || m->p.status >= 700) {
      error("userapi_handle_response",
            "Response with surprising status %d received\n", m->p.status);
      m->handled = finished_handling;
    }
    else if (m->p.status >= 100 && m->p.status < 700) {
      /* We received a response from API. */
      /* Send it immediately */
      if (m->type == message_real_message) {
        debug(DEBUG_MISC, "user_handle_response",
              "calling policy_forward_user_response\n");
        policy_forward_user_response(r, m);
      }
      else if (m->type == message_pseudo_message) {
        debug(DEBUG_MISC, "user_handle_response", "calling policy_send_new_response\n");
        policy_send_new_response(r, m->p.status, m->p.reason);
      }
      else {
        error("userapi_handle_response","surprising message type %d\n",
          m->type);
      }

      m->handled = finished_handling;
    }

    if (r->sent_200) {
      info->state = user_info::final_wait;
    }
  }
  else if (strcmp(r->p.method, "REGISTER") == 0) {
    /* This is response to register */
    /* This response is from remote for an outgoing request */
    sipreg_t *reg = NULL;
    debug(DEBUG_MISC, "userapi_handle_response", "Response %d for REGISTER\n",
          m->p.status);

    reg = SipRegSearch(info->callid);
    if (reg == NULL) {
      error("userapi_handle_response", "No reg call present\n");
      m->handled = finished_handling;
      return transaction_policy_ok;
    }

    if (m->p.status < 100 || m->p.status >= 700) {
      error("userapi_handle_response",
            "Response with surprising status %d received\n", m->p.status);
      m->handled = finished_handling;
    }
    else if (m->p.status >= 200 && m->p.status < 700) {
      /* We received a 2xx response. */
      /* Send it immediately, and cancel all other branches. */
      /* Hand it up to the call thread. */
      MessageEnqueue(&reg->mqueue, m->type, &m->p,
                     &m->body, &m->source, 
                     NULL,
                     nobody_handled);
      pthread_cond_broadcast(&reg->cond);

      /* Should we set sent_final and sent_200 to 1 here XXX ? */
      r->sent_final = 1;
      if (m->p.status < 300) {
        r->sent_200 = 1;
      }

      info->state = user_info::final_wait; /* XXX is this correct ? */

      policy_set_timeout(r, (double)config.timeout, transaction_policy);
      m->handled = finished_handling;
    }
    pthread_mutex_unlock(&reg->mutex);
  }
  /* Response Handling for MESSAGE */
  else if (strcmp(r->p.method, "MESSAGE") == 0) {
    /* This response is from remote for a MESSAGE sent */
    if (m->p.status >= 200 && m->p.status < 700) {
      debug(DEBUG_MISC, "userapi_handle_response", "Received the response (%d) for (%s)\n",
                                                    m->p.status, info->callid);
      m->handled = finished_handling;

      r->done = 1;
      policy_done(r, transaction_policy);

      return transaction_policy_ok;
    }
  }
  else {
    /* This response is from remote for an outgoing request */
    call = CallSearch(info->callid);
    if (call == NULL) {
      error("userapi_handle_response", "No call present (with callid %s)\n",
            info->callid ? info->callid : "NULL");
      m->handled = finished_handling;
      return transaction_policy_ok;
    }

    if (m->p.status < 100 || m->p.status >= 700) {
      error("userapi_handle_response",
            "Response with surprising status %d received\n", m->p.status);
      m->handled = finished_handling;
    }
    else if (m->p.status >= 200 && m->p.status < 300) {
      /* Get the tag from the request an add it to the 
       * call data structure */      
      parser_header_t *h;
      if ((h = RFC822_Header(m->p.headers, "To")) && h->elements) {
        /* Update the tag */

        char *tag = RFC822_Attribute1(h->elements->parameters, "tag");
        if (tag == NULL) {
          /* debug, no tag found */
          debug(DEBUG_MISC, "userapi_handle_response", "No tag in To header, fine.\n");
          /* should we remove the call->remote_tag here ? 
           * I don't think, based on the From, To comparision 
           * method. */
        }
        else {
          if (call->remote_tag) {
            /* Tag was already received in a previous message */
            if (!strcmp(call->remote_tag, tag)) {
              /* Tags match, fine */
            }
            else {
              /* Tag mismatch */
              error("userapi_handle_response", 
                "Tag mismatch in To header, overwriting\n");
            }
          }
          /* Just copt the tag to call data structure */
          strassign(&call->remote_tag, tag);
        }
      }

      /* Update the tag in r if not already there */
      if ((h = RFC822_Header(r->p.headers, "To")) && h->elements) {
        char *tag = RFC822_Attribute1(h->elements->parameters, "tag");
        if (tag == NULL && call->remote_tag) {
          /* store the tag in r */
          RFC822_AddAttribute(h, "tag", call->remote_tag, PT_token);
        }
      }
      
      /* We received a 2xx response. */
      /* Send it immediately, and cancel all other branches. */
      /* Hand it up to the call thread. */
      MessageEnqueue(&call->mqueue, message_real_message, &m->p,
                     &m->body, &m->source, 
                     NULL,
                     nobody_handled);
      pthread_cond_broadcast(&call->cond);

      /* Should we set sent_final and sent_200 to 1 here XXX ? */
      r->sent_final = 1;
      if (m->p.status < 300) {
        r->sent_200 = 1;
      }

      info->state = user_info::final_wait; /* XXX is this correct ? */


      /* If we get multiple 200's, we may cancel more than once here; but
       * that's okay. */
      CancelUnfinishedBranches(r);

      policy_set_timeout(r, (double)config.timeout, transaction_policy);
      m->handled = finished_handling;
    }
    else {
      /* If we've already sent a final response, don't bother to send or keep
       * any non-2xx responses. */
      if (r->sent_final) {
        m->handled = finished_handling;
      }
      else {
        if (m->p.status < 200) {
          /* 1xx response received.  If non-100, give it to call thread */
          if (m->p.status != 100) {
            MessageEnqueue(&call->mqueue, message_real_message, &m->p,
                           &m->body, &m->source, 
                           NULL,
                           nobody_handled);      
            pthread_cond_broadcast(&call->cond);
          }
          m->handled = finished_handling;
        }
        /* 2xx responses already handled above */
        else if (m->p.status < 400) {
          /* We received a 3xx response.  If ProxyRecursion is turned on,
           * try to do 3xx recursion.  If not, or if that function says it
           * can't do any recursion on this response, keep it in the message
           * queue for later consideration. */ 
          if (config.ProxyRecursion && Do3xxRecursion(r, m)) {
            m->handled = finished_handling;
          }
          else {
            m->handled = policy_handled;
          }
        }
        else if (m->p.status >= 400) {
          /* 4xx - 6xx response received */
          /* Keep this message around for later decisions of which response to
           * send. */
          m->handled = policy_handled;
        }

        if (r->branches_outstanding <= 0) {
          /* send the best response to the API */
          NextUserProxyStep(r, call); /* choose next step */
        }
      }
    }

    pthread_mutex_unlock(&call->mutex);
  }
  return transaction_policy_ok;
}

/* NextUserProxyStep()
 * This is called when it's time to do the "next step" of proxying -- either
 * a timeout has expired, or all currently active proxy branches have
 * returned a final response.  If we have more branches we can still try to
 * start, and we haven't received a 200, do so; otherwise, pick and send the
 * best response to the call thread. This is called for
 * outgoing requests only.
 */
static void NextUserProxyStep(request_t *r, sipcall_t *call)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;
  message_t *best;

  switch (info->state) {
  case user_info::sending_requests:
    info->more = StartSomeRegBranches(r, info->reg, &info->pos);
    if (!info->more) {
      info->state = user_info::awaiting_responses;
    }
    policy_set_timeout(r, (double)config.timeout, transaction_policy);
    break;
  case user_info::awaiting_responses:
    /* This function is defined in proxy.c */
    best = PickBestResponse(r, r->mqueue);
  
    if (r->sent_final) {
      /* Sent by PickBestResponse.  Do nothing. */
      ;
    }
    else if (best == NULL) {
      int status;
      /* If no responses arrived, it means the response timed out.
       * If some responses arrived, but we didn't find any, it means
       * we (e.g.) had two 300-class responses which pointed to each other.
       * The status is 408 in the former case, and 404 in the latter.
       */
      if (r->branches_outstanding == r->branch_count) {
        status = 408;
      }
      else {
        status = 404;
      }

      r->sent_final = 1;

      /* Hand it up to the call thread. */
      MessageEnqueuePseudoResponse(&call->mqueue, 
                                  NULL,
                                  nobody_handled, status);
      pthread_cond_broadcast(&call->cond);
    }
    else if (best->type == message_pseudo_message) {
      r->sent_final = 1;
      if (best->p.status < 300) {
        r->sent_200 = 1;
      }

      MessageEnqueuePseudoResponse(&call->mqueue, 
                                  NULL,
                                  nobody_handled, best->p.status);
      pthread_cond_broadcast(&call->cond);
    }
    else {
      r->sent_final = 1;
      if (best->p.status < 300) {
        r->sent_200 = 1;
      }

      MessageEnqueue(&call->mqueue, message_real_message, &best->p,
                     &best->body, &best->source, 
                     NULL,
                     nobody_handled);
      pthread_cond_broadcast(&call->cond);
    }
    /* Once we send the response, go to final wait state. */
    info->state = user_info::final_wait;
    policy_set_timeout(r, (double)config.timeout, transaction_policy);
    break;

  default:
    error("NextUserProxyStep", "called with proxy in inappropriate state %d",
          info->state);
    break;
  }
} /* NextProxyStep */


/* userapi_timeout()
 * Called when a timeout expires for the proxy
 * If we're sending requests, call NextProxyStep; if we're
 * awaiting responses, just refresh the timeout (the clients will time out
 * and send up a pseudo-response); if we're in final wait, we're done.
 */
static transaction_policy_ret userapi_timeout(request_t *r)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;
  sipcall_t *call = NULL;

  if (info->reg == NULL) {
    /* This is an incoming request, should this be handled, XXX */
    debug(DEBUG_MISC, "userapi_timeout", "Timeout for incoming request not handled\n");
    switch (info->state) {
    case user_info::sending_requests:
      policy_done(r, transaction_policy); /* terminate the policy */
      break;
    case user_info::awaiting_responses:
      policy_set_timeout(r, (double)config.timeout, transaction_policy);
      break;
    case user_info::final_wait:
      policy_done(r, transaction_policy); /* timedout in final_wait */
      break;
    default:
      debug(DEBUG_MISC, "userapi_timeout", "Returning error as state = %d\n",
            info->state);
      return transaction_policy_ok;
      break;
    }    
  }
  else {
    /* This is an outgoing request */
    switch (info->state) {
    case user_info::sending_requests:
      call = CallSearch(info->callid);
      if (call == NULL) {
        /*
         * Should we process it here also; 
         * similar to awaiting_responses state.
         * for forking REGISTER XXX 
         */
        error("userapi_timeout", "No call present\n");
        policy_done(r, transaction_policy); /* terminate the policy */
      }
      else {
        NextUserProxyStep(r, call); /* choose next step */
        pthread_mutex_unlock(&call->mutex);
      }
      break;
    case user_info::awaiting_responses:
      /* We call the NextUserProxyStep, unlock proxy.c
       * when the timeout expires in awaiting_responses state also...
       */
      call = CallSearch(info->callid);
      if (call == NULL) {
        /* This response is from remote for an outgoing request */
        sipreg_t *reg = NULL;
        reg = SipRegSearch(info->callid);
        if (r->p.method != NULL
             && !strcmp(r->p.method, "REGISTER")
             && reg != NULL) {
          /* We received a 2xx response. */
          /* Send it immediately, and cancel all other branches. */
          /* Hand it up to the call thread. */
          MessageEnqueuePseudoResponse(&reg->mqueue, 
                                       NULL,
                                       nobody_handled,
                                       404 /* Not found */);
          pthread_cond_broadcast(&reg->cond);
          
          /* Should we set sent_final and sent_200 to 1 here XXX ? */
          r->sent_final = 1;
          
          info->state = user_info::final_wait; /* XXX is this correct ? */
          
          policy_set_timeout(r, (double)config.timeout, transaction_policy);
        }
        else {
          error("userapi_timeout", "No reg/call present\n");
          policy_done(r, transaction_policy); /* terminate the policy */
        }

        /* Unlock the reg object */
        if (reg != NULL) {
          pthread_mutex_unlock(&reg->mutex);
        }
      }
      else {
        NextUserProxyStep(r, call); /* choose next step */
        pthread_mutex_unlock(&call->mutex);
      }
      //policy_set_timeout(r, (double)config.timeout, transaction_policy);
      break;
    case user_info::final_wait:
      policy_done(r, transaction_policy); /* timedout in final_wait */
      break;
    default:
      return transaction_policy_error;
      break;
    }
  }

  return transaction_policy_ok;
}


/* userapi_cleanup()
 * Called when cleaning up the userapi policy
 * Free userapi-specific memory
 */
static transaction_policy_ret userapi_cleanup(request_t *r)
{
  struct user_info *info = (struct user_info *)r->transaction_policy_info;

  free_valid(info->callid);
  if (info->reg) {
    /* policy_info_t should be freed in proxy.c also ? XXX */
    //XXX free_valid(info->reg->script);
    SIP_FreeContact(info->reg->contact);
    //XXX free_valid(info->reg->content_type);
  }
  free_valid(info->reg);
  free_valid(r->transaction_policy_info);
  return transaction_policy_ok;
} /* userapi_cleanup */

/* --------------------------- following C definitions are needed -------------- */

#ifdef __cplusplus
extern "C" {
#endif

transaction_policy_t userapi_policy =
{
  "userapi",
  &userapi_init,
  &userapi_handle_initial_request,
  &userapi_handle_subsequent_request,
  &userapi_handle_response,
  &userapi_timeout,
  &userapi_cleanup
};

/* Following definition is needed by sip library */

const int missing_content_length_means_zero = 0;

int SIP_HandleRequestLocally(method_type_t method, uri_t uri)
{
  return 1; /* Always local, since no PROXY yet. XXX */
}

#ifdef __cplusplus
}
#endif
