/*
 * asr.c: The thread defined here receives packets over RTP/RTCP,
 *        does silence detection, and automatic speech recognition
 *        using CMU Sphinx2.
 *
 * Copyright 2002 by Columbia University; all rights reserved
 */

#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <pthread.h>
#include "time.h"
#include "simv.h"

extern "C" {
int16 pcmu_to_l16(unsigned int ulawbyte);
#include <sphinx2/fbs.h>
#include <sphinx2/ad.h>
}


/*
 * Caculate the SNR value
 * Excerpted from "radio.cpp" made by Kundan
 */
#define SQR(x) (x)*(x)
static double EnergyLevel (unsigned char *dataBuf, int dataLen) 
{
  int i, n;
  double sum_x, avg_x;

  n= dataLen;

  /* Following portion of the code is copied from HW2 */
  for (sum_x=0, i=0; i < n; i++) {
    sum_x += pcmu_l16(dataBuf[i]);
  }
  avg_x = sum_x/(double)n;

  for (sum_x=0, i=0; i < n; i++) {
    sum_x += SQR(pcmu_l16(dataBuf[i]) - avg_x);
  }
  avg_x = sum_x/(double)n;

  return (avg_x);

} /* Energy Level */


/*
 * AsrRtpRecvThread()
 *
 * Main ASR thread.
 */
void *AsrRtpRecvThread(void *_args)
{

  /* Argument from the parent */
  simv_info_t *asr = (simv_info_t *)_args;

  /* For pcmu -> pcml16 conversion */
  int16 l16_data[RTP_RECV_SIZE];

  /* Voice related */
  int rcnt, sil_cnt, sil_total, utt_total, seq;
  bool talkspurt, sound;
  double el;
  int sil_dur_byte = SILENCE_DURATION_SEC * 8000;
  int max_utt_byte = MAX_UTT_SEC * 8000;

  /* RTP related */
  char recv_buf[RTP_RECV_SIZE];
  int loaduse;    
  socktype rtpsock, rtcpsock;
  rtp_packet rtppkt;
  rtperror the_err;
  struct timeval socktv;
  fd_set selection;
  int maxfd = 0;
  int nopacket = 0;

  /* ASR related */
  char *hyp, tmp_msg[MAX_ASR_LENGTH];
  int32 asr_err, frm;

  /* MISC */
  int i;

  /* Check the arguments */
  if (asr == NULL) {
    fprintf( stderr, "Error:AsrRtpRecvThread: simv_info_t is NULL\n");
    return NULL;
  }
  debug(DEBUG_MISC,"AsrRtpRecvThread","Started ASR Thread (VAT:%10.2f)\n", asr->vat_th);

  /* If you want sphinx to log voice data */
  /* uttproc_set_rawlogdir("./asrlog");   */

  /* Get RTP/RTCP socket */
  the_err = RTPSessionGetRTPSocket(asr->rtp_context, &rtpsock);
  if( the_err != RTP_OK){
    fprintf( stderr, "Error:AsrRtpRecvThread: RTPSessionGetRtpSocket: %s\n",
             RTPStrError(the_err));
    return NULL;
  }
  
  the_err = RTPSessionGetRTCPSocket(asr->rtp_context, &rtcpsock);
  if( the_err != RTP_OK){
    fprintf( stderr, "Error:AsrRtpRecvThread: RTPSessionGetRtcpSocket: %s\n",
             RTPStrError(the_err));
    return NULL;
  }

  /* max fds for select */
  if (rtpsock < rtcpsock) {
    maxfd = rtcpsock;
  }
  else {
    maxfd = rtpsock;
  }

  /* ASR Main Loop */
  while(!asr->oAsrTerminate)
  {
    uttproc_begin_utt(NULL);

    talkspurt = false;
    utt_total = sil_cnt = sil_total = seq =0;

    nopacket = 0;

    /**
     *  Utterance main loop. 
     *  
     *  Exits the loop either when it has passed
     *  SILENCE_DURATION_SEC (1 sec) after any sound packet
     *  or the accumulated sound reached to the max. length,
     *  which happens if the speaker keeps talking for
     *  MAX_UTT_SEC without silence longer than
     *  SILENCE_DURATION__SEC.
     *
     *  Then proceed to ASR processing.
     */
    while(utt_total < max_utt_byte){
      int e;

      FD_ZERO(&selection);
      FD_SET(rtpsock, &selection);
      FD_SET(rtcpsock, &selection);
  
      socktv.tv_sec = 1;
      socktv.tv_usec = 0; 

      /* Recv UDP packets from RTP/RTCP socket */
      if ( (e = select(maxfd + 1, &selection, NULL, NULL, &socktv)) > 0 ) {

        if ( FD_ISSET(rtcpsock, &selection)) {
          /* Receive data over RTCP */
          loaduse = RTP_RECV_SIZE;
          
          pthread_mutex_lock(&asr->rtp_mutex);
          the_err = RTPReceive(asr->rtp_context,
                               rtcpsock,
                               recv_buf,
                               &loaduse);
          pthread_mutex_unlock(&asr->rtp_mutex);
          /* Ignore the RTCP packet */
        }
        
        if ( FD_ISSET (rtpsock, &selection)) {
          /* Receive voice data over RTP */
          loaduse = RTP_RECV_SIZE;
          pthread_mutex_lock(&asr->rtp_mutex);
          the_err = RTPReceive(asr->rtp_context, rtpsock, recv_buf, &loaduse);
          pthread_mutex_unlock(&asr->rtp_mutex);

          nopacket = 0;

          /* Extract packet */
          rtppkt = RTPGetRTPPacket(recv_buf, loaduse);

          /*
           * Silence Detection:
           * Assummig we're receiving pcmu from sipc,
           * The default threshold is defined by VAT_THRESHOLD
           * or it be specified in the argument.
           */
          rcnt = rtppkt.payload_len;
          el = EnergyLevel( (unsigned char *)rtppkt.payload, rcnt);

          if( el > asr->vat_th)
            sound = true;
          else
            sound = false;

          /** 
           * Processing POLICY:
           * A voice slot length is defined by MAX_UTT_SEC.
           * The silence in the begining of a slot is discarded,
           * but once sound is detected, silences between words are
           * accumulated into Sphinx ASR engine. If the silence is
           * longer than SILENCE_DURATION_SEC, that will regarded as
           * the end of a talkspurt. Then proceed to ASR process.
           */ 
          if(sound){ /* Sound data */
            if(!talkspurt)
              talkspurt = true;
              debug(DEBUG_MISC,"AsrRtpRecvThread",
                    "START of a Talkspurt:(VAT=%10.2f)\n", asr->vat_th);
    
            /* Reset silence timer */
            sil_cnt = 0;
    
          }
          else { /* Silence data */
            /*
             * There are two kinds of silence here.
             * One is in a talkspurt, which should be accumulated
             * into ASR engine,  the other one is in a non-talkspurt,
             * which should be discarded.
             */
            if(talkspurt){
              /* Increment silence timer */
              sil_cnt += rcnt;
    
            } else {
              /* Count silence in non-talkspurt */
              sil_total += rcnt;
    
              /* No sound was detected within the slot */
              if(sil_total > max_utt_byte)
                break;
              else 
                continue; /* Go get next packet */
            }
          }

          /*
           * The end of the talkspurt, the silence after
           * sound lasted for SILENCE_DURATION_SEC.
           * Let's proceed to ASR process.
           */
          if(sil_cnt > sil_dur_byte) {
            debug(DEBUG_MISC,"AsrRtpRecvThread",
                  "END of a Talkspurt:(Silence %d bytes, %d sec)\n",
                   sil_cnt, SILENCE_DURATION_SEC);
            break;
          }
    
          /*  Convert pcmu to liner 16 bits */
          for(i=0; i<rcnt; i++)
            l16_data[i] = pcmu_to_l16((unsigned int)rtppkt.payload[i]); 
    
          /* Send data to Sphinx engine */
          uttproc_rawdata(l16_data, rcnt, 0);
          utt_total += rcnt; seq++;
    
        } /* RTP packet reeceived */

      } /* select() returns > 0 */
      else if (e == 0) {
        /* No packet came */
        nopacket ++;
        if (nopacket >= SILENCE_NOTIFY_SEC) { 
          /*
           * No packet received for SILENCE_NOTIFY_SEC seconds.
           * then break from inner loop and inform user of
           * of this long silence. This happens if sipc does
           * silence packet supression.
           */
          break;
        }
      }  

      /* Check if we should be terminated here */
      if(asr->oAsrTerminate)
        break;
      
    } /* inner while loop */
    
    /* Check if we should be terminated again */
    if(asr->oAsrTerminate)
      break;
    
    /** Sphinx2 ASR Processing */
    if(talkspurt){
      uttproc_end_utt();
      asr_err = uttproc_result(&frm, &hyp, 1);
      if(asr_err < 0){
        fprintf(stderr, "Error:AsrRtpRecvThread: uttproc_result()\n");
        return NULL;
      }

      debug(DEBUG_MISC,"AsrRtpRecvThread",
            "Result: %d bytes, %3.1f sec, frm=%d, ASR=[%s]\n",
            utt_total, utt_total/8000.0, frm, hyp);

      if(strlen(hyp) <= 1)
        /* Couldn't be recognized */
        strcpy(tmp_msg, "Unrecognizable Speech!!");
      else
        strcpy(tmp_msg, hyp);
      
    } else {
      /* Clean up buffer in Sphinx engine */
      uttproc_abort_utt();

      debug(DEBUG_MISC,"AsrRtpRecvThread",
            "Silence: %d bytes %3.1f sec, NOT ASRed\n",
            sil_total, sil_total/8000.0);

      /* Let's inform the application of long silence */
      sprintf(tmp_msg, "Silence %d sec!!", SILENCE_NOTIFY_SEC);
    }
    
    /* Callback app. level function to send IM */
    asr->asr_mesg = strdup(tmp_msg);
    debug(DEBUG_MISC,"AsrRtpRecvThread","Calling OnEndTalkSpurt()\n");
    OnEndTalkSpurt(asr);
    
    free(asr->asr_mesg);
    
  } /* while thread is active */

  /** Delete the instance */
  debug(DEBUG_MISC,"AsrRtpRecvThread","ASR Thread exitting normally.\n");
  pthread_exit((void *) 0);

} /* AsrRtpRecvThread() */


