/*
 * main.cpp - Main file for SImVoice
 * SIP IM/Voice Converter
 * (c) 2002 Columbia University. All rights reserved.
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>

#include "iptelapi.h"
#include "sipapi.h"

extern "C" {
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "config.h"
#include "sysdep.h"
#include "strdupn.h" /* free_valid */
#include "base64.h" 

#include <sphinx2/fbs.h>
extern char *optarg;
extern u_long cine_random32(int);
}

#include "pthread.h"
#include "pthread-util.h"
#include "simv.h"

/**
 * MySIPUserAgent and MySIPCall
 */
class MySIPUserAgent : public SIPEndpoint
{

public:
  MySIPUserAgent(bool IsServer = true, bool IsRegistrar=false);
  MySIPUserAgent(const IptelIPAddress& ip4,
                 const IptelIPAddress& ip6, bool IsRegistrar=false);
  virtual IptelCall *AllocateNewCallObject(const char *callid);

  /* SIP MESSAGE Handler */
  virtual int OnIncomingSIPMessage(const SIPInstantMessageInfo & iminfo);

public:
  /** IM-Voice Session Table Handlers */
  simv_info_t *SearchImvsBySid(const char *sid);
  simv_info_t *SearchImvsByCid(const char *callid);
  void InsertImvs(simv_info_t *imvs);
  void DeleteImvs(simv_info_t *imvs);

protected:
  /** Access point of the IM session table */
  simv_info_t *imvs_tab;
};

class MySIPCall : public SIPCall
{
public:
  MySIPCall(const char *callid=NULL) : SIPCall(callid) { };

  MySIPCall(SIPEndpoint& ep, const char *callid = NULL) : 
         SIPCall(ep,callid) { };

  virtual ~MySIPCall() { };

  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();
};

/* SImVoice server itself */
static MySIPUserAgent *sip = NULL;

/* Call to be processed */
static MySIPCall *cur_call = NULL;

/* Im-Voice session to be processed */
static simv_info_t *cur_imvs = NULL; 

/* Simvoice status */
static int simv_status = SIMV_NORMAL;

/* Silence detection threshold */
static double vat_th = VAT_THRESHOLD;

GenericHeaders globals;

/*
 * SendVoiceMessage()
 * Initializes TTS thread with text data,
 * the thread will send synthesized voice.
 * over RTP.
 */
static void SendVoiceMessage(simv_info_t *myimvs, char *mesg_str)
{
  /** Initiating TTS Thread */
  pthread_attr_t attr;
  int oRet, status;

  /** Making attribute for a joinable thread */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE); 

  /** MESSAGE to be synthesized */
  myimvs->tts_mesg = mesg_str;

  /* XXX: Create a thread in joinable state */
  oRet = pthread_create(&myimvs->tts_thread, &attr, TtsRtpSendThread, myimvs);
  if(oRet)
  {
    fprintf(stderr, "simvoice>> SendVoice(ERROR): pthread_create(TTS)\n");
    exit(1);
  }
  pthread_attr_destroy(&attr);

  /** XXX: Waiting for the thread to terminate */
  oRet = pthread_join(myimvs->tts_thread, (void **)&status);
  if(oRet)
  {
    fprintf(stderr, "simvoice>> SendVoice(ERROR): pthread_join(TTS)\n");
    exit(1);
  }
  debug(DEBUG_MISC,"simvoice","TTS Thread completed\n");  

} /* SendVoiceMessage() */


/** Methods for the class MySIPUserAgent */

MySIPUserAgent::MySIPUserAgent(bool IsServer, bool IsRegistrar)
: SIPEndpoint(IsServer,IsRegistrar)
{
  imvs_tab = NULL;
  /* Do nothing here */
}

MySIPUserAgent::MySIPUserAgent(const IptelIPAddress& ip4, 
                               const IptelIPAddress& ip6, bool IsRegistrar)
: SIPEndpoint(ip4, ip6, IsRegistrar)
{
  imvs_tab = NULL;
  /* Do nothing here */
}

IptelCall *MySIPUserAgent::AllocateNewCallObject(const char *callid)
{
  /* Allocate a new SIP call */
  return new MySIPCall(*this, callid);
}

/*
 * MySIPUserAgent::OnIncomingSIPMessage(const SIPInstantMessageInfo & iminfo)
 *
 * Incoming message handler
 */
int MySIPUserAgent::OnIncomingSIPMessage(const SIPInstantMessageInfo & iminfo){

  /**
   *  Search Existing IM-Voice session
   *  Associated with this MESSAGE first.
   */
  const char *sid;
  simv_info_t *myimvs = NULL;

  sid = DStringValue((DString*)&iminfo.sid);
  debug(DEBUG_MISC,"simvoice","Incoming IM sid=[%s] cont=[%s]\n",
        sid, iminfo.imbody);

  /** Search Im-Voice session */
  myimvs = sip->SearchImvsBySid(sid);

  /** If associated session couldn't be found */
  if(myimvs == NULL){
    debug(DEBUG_MISC,"simvoice","No associated session found sid=[%s]!!\n", sid);

    /* Inform IM user of the error */
    SIPInstantMessageInfo res_iminfo;
    char mesg[MAX_IM_LENGTH];

    sprintf(mesg, "\n*** [SImVoice@%s] ***\n"
                   " Sorry, no associated voice session found!!\n"
                   " Following your message couldn't be delivered.\n"
                   " [%s]\n",
                   (const char *)SIPLibrary::HostName4,
                    iminfo.imbody);
    res_iminfo.imbody = mesg;
    res_iminfo.content_length = strlen(mesg);

    DStringInit(&res_iminfo.sid);
    DStringAppendD(&res_iminfo.sid, &iminfo.sid);

    IptelAddress to, from;
    to   = iminfo.cnt;
    from = iminfo.dst;

    debug(DEBUG_MISC,"simvoice","Sending IM: to=[%s], cnt=[%s]",
         (const char *)to, res_iminfo.imbody);

    sip->SIPSendMessage(to, from, res_iminfo);
    return(0);
  }

  /** Send TTS voice to the caller */
  SendVoiceMessage(myimvs, iminfo.imbody);

  return(0);
} /* OnIncomingSIPMessage() */


/*
 * MySIPUserAgent::SearchImvsBySid(const char *sid)
 * 
 * Search session table using sid (session id)
 */
simv_info_t *MySIPUserAgent::SearchImvsBySid(const char *sid)
{
  simv_info_t *cur;
  const char *cur_sid;

  for(cur=imvs_tab; cur != NULL; cur = cur->next){
    cur_sid = DStringValue(&cur->sid);
    if(strcmp(sid, cur_sid) == 0)
      return cur;
  } 

  return NULL;
} /* SearchImvsBySid() */


/*
 * MySIPUserAgent::SearchImvsByCid(const char *callid)
 * 
 * Search session table using cid (call id)
 */
simv_info_t *MySIPUserAgent::SearchImvsByCid(const char *callid)
{
  simv_info_t *cur;
  const char *cur_cid;

  for(cur=imvs_tab; cur != NULL; cur=cur->next){
    cur_cid = DStringValue(&cur->cid);
    if(strcmp(callid, cur_cid) == 0)
      return cur;
  } 
  return NULL;
} /* SearchImvsByCid() */


/*
 * MySIPUserAgent::InsertImvs(simv_info_t *imvs)
 * 
 * Insert session into table (linked list)
 */
void MySIPUserAgent::InsertImvs(simv_info_t *imvs)
{
  simv_info_t *cur;

  imvs->next = NULL;

  /** The first entry */
  if(imvs_tab == NULL){
    imvs_tab = imvs;
    return;
  }

  for(cur=imvs_tab; cur->next != NULL; cur=cur->next)
    ;
  cur->next = imvs;

} /* InsertImvs() */


/*
 * MySIPUserAgent::DeleteImvs(simv_info_t *imvs)
 * 
 * Delete session from table (linked list)
 */
void MySIPUserAgent::DeleteImvs(simv_info_t *imvs)
{
  simv_info_t *cur, *prev;
  const char *cid, *cur_cid;

  cid = DStringValue(&imvs->cid);
  cur = prev = imvs_tab;

  while(cur != NULL){
    cur_cid = DStringValue(&cur->cid);
    if(strcmp(cid, cur_cid) == 0){

      /** First entry */
      if(cur == prev){
        imvs_tab = cur->next;
      }
      prev->next = cur->next;
      DStringFree(&cur->cid);
      DStringFree(&cur->sid);
      delete cur;
      return;
    }
    prev = cur;
    cur  = cur->next;
  } 
} /* DeleteImvs() */


///// Methods for the class MySIPCall /////

/*
 * MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
 * 
 * Handles new incoming call
 */
void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  IptelAddress dest    = GetSelfAddress();
  IptelAddress src     = GetRemoteAddress();

  IptelAddress im_user;
  IptelAddress server;
  int isv6 = 0;

  /*
   * Set the global variable for call.
   * Accept or reject is called in UI thread 
   */
  cur_call = this;

  debug(DEBUG_MISC,"simvoice","Incoming call from: %s\n" ,(const char *)src);

  /*
   * Always accept an incoming call
   */
  MySIPCall *mycall;
  mycall = this;
  if (mycall == NULL) {
    cerr << "simvoice>> Error: Call pointer is NULL" << endl;
    return;
  }

  /**
   * Naoya Seta
   * Start RTP library tool
   */ 
  rtperror the_err; 

  /** Create new IM-Voice session */
  simv_info_t *imvs = new simv_info_t;

  imvs->vat_th = vat_th;
  pthread_mutex_init(&(imvs->rtp_mutex), NULL);

  the_err = RTPCreate(&imvs->rtp_context);
  debug(DEBUG_MISC, "OnNewIncoming:"," Created RTP context %d\n",
        (int)imvs->rtp_context);
  if ( the_err != RTP_OK ) {
    fprintf(stderr, "ERROR:  OnNewIncoming:  CreateRTP Failed\n");
    exit(1);
  }
  
  /* Set the default receive address, INADDR_ANY, Dyanamic port */
  the_err = RTPSessionSetReceiveAddr(imvs->rtp_context, NULL, 0 );
  if ( the_err != RTP_OK ) {
    fprintf(stderr, "ERROR: RTPSessionSetReceiveAddr Failed\n");
    exit(1);
  }

  /* --- Setting Sending Address  --- */
  SessionDescription *rem_session;
  rem_session = mycall->GetRemoteSessionDescription();
  MediaInfo rem_media;  
  IptelIPAddress rem_addr;
  struct sockaddr_storage addr;
  
  if (rem_session != NULL) {
    if (rem_session->FindSession(rem_media, MediaInfo::Receive, 
                             MediaInfo::Audio)) {
      rem_addr = rem_media.GetRxAddress();
      rem_addr.GetIpAddress(addr);
      AFISetPort(&addr, rem_media.GetRxPort());
      if (addr.ss_family == AF_INET6) {
        isv6 = 1;
      }
    }
  }
  else {
    cerr << "ERROR: No remote session in INVITE" << endl;
  }

  /* Add sending address to RTP */
  the_err = RTPSessionAddSendAddr(imvs->rtp_context, 
	      AFIGetDottedAddress((const struct sockaddr*) &addr),
	      AFIGetPort(&addr), 255);

  if (the_err != RTP_OK) {
    fprintf(stderr, "ERROR: RTPSessionAddSendAddr: %s\n", RTPStrError(the_err));
  }
  else {
    debug(DEBUG_MISC, "Added RTPSend Address", " %s : %d\n",
      AFIGetDottedAddress((const struct sockaddr*) &addr), AFIGetPort(&addr));

    /* Set a CNAME */
    the_err = RTPMemberInfoSetSDES(imvs->rtp_context, 0, RTP_MI_CNAME, "simvoice@localhost");
    if ( the_err != RTP_OK ) {
      fprintf(stderr, "ERROR: RTPMemberINfoSetSDES: %s\n", RTPStrError(the_err));
    }

    /* NAME */
    the_err = RTPMemberInfoSetSDES(imvs->rtp_context, 0, RTP_MI_NAME, "SIMVOICE");
    if ( the_err != RTP_OK ) {
      fprintf(stderr, "ERROR: RTPMemberINfoSetSDES: %s\n", RTPStrError(the_err));
    }
      
    /* Open the RTP Connection (Just creates receiving socket) */
    the_err = RTPOpenConnection(imvs->rtp_context);
    if ( the_err != RTP_OK ) {
      fprintf(stderr, "ERROR: RTPOpenConnection: %s\n", RTPStrError(the_err));
    }

    /* --- Allocate self session info to accept the call--- */
    SessionDescription *self_session;
    self_session = new SessionDescription;
  
    MediaInfo media;
    media.SetDirection(MediaInfo::Receive);
    media.SetDataType(MediaInfo::Audio);
    FormatInfo f(0);
    media.AddFormat(f);

    char address[1024];
    unsigned short rtp_port;
	
    /* This must be called after RTPOpenConnection  */
    the_err = RTPSessionGetReceiveAddr(imvs->rtp_context,
			 address, &rtp_port);
    if(the_err != RTP_OK){
      fprintf(stderr, "ERROR: RTPSessionGetReceiveAddrr: %s\n", RTPStrError(the_err));
    }
      
    if (AFIIPv6Active() && isv6) {
      media.SetRxAddress(IptelIPAddress(SIPLibrary::HostInAddr6));
      media.SetAddressType(MediaInfo::AddressIP6);
      media.SetRxPort(rtp_port);
      self_session->AddSession(media);
    }
    else {
     /* self IP address */
     media.SetRxAddress(IptelIPAddress(SIPLibrary::HostInAddr4));
     media.SetRxPort(rtp_port);
     self_session->AddSession(media);
    }

    mycall->SetSelfSessionDescription(self_session);
  }

  /* Send out an initial MESSAGE to the callee */

  /** Decode base64 encoded final destination */
  im_user = base64_decode_string(dest.GetUser());

  /* XXX: Correct? */
  im_user.SetUser(im_user.GetUser());
  im_user.SetScheme("sip");
  
  /** Set Im-Voice id (currently call id) */
  DStringInit(&imvs->sid);
  DStringInit(&imvs->cid);
  DStringPrintf(&imvs->sid, "%lu", SIPLibrary::Random32());
  DStringPrintf(&imvs->cid, "%s", GetCallId());

  server = SIPLibrary::HostInAddr4;
  server.SetUser(strdup(DStringValue(&imvs->sid)));
  server.SetScheme("sip");

  SIPInstantMessageInfo iminfo;
  char mesg[MAX_IM_LENGTH];

  sprintf( mesg, "\n*** [SImVoice@%s] ***\n"
                 "  You've got a message from:\n"
                 "  [%s]\n",
                  (const char *)SIPLibrary::HostName4,
                  (const char *)src.GetURI());
  iminfo.imbody = mesg;
  iminfo.content_length = strlen(mesg);
  DStringInit(&iminfo.sid);
  DStringAppendD(&iminfo.sid, &imvs->sid);

  debug(DEBUG_MISC,"simvoice", "Sending IM:\n from=[%s] to=[%s]\n cnt=[%s]\n",
        (const char *)server, (const char *)im_user, mesg);

  /** XXX: Can I copy like this? */
  imvs->im_user     = im_user;
  //imvs->voice_user  = src;
  imvs->voice_user  = GetIncomingRequestHeader("Contact");
  imvs->server      = server;
  cur_imvs          = imvs;

  int ret;
  ret = sip->SIPSendMessage(im_user, server, iminfo);

  if(ret > 0){
    /** IM destination was unreachable */
    simv_status = SIMV_REJECT_PEND;
    debug(DEBUG_MISC,"simvoice","IM unreachable [%s]\n", (const char *)im_user);

  } else {
    /** Succeed to send IM */
    simv_status = SIMV_ACCEPT_PEND;
  }

} /* OnNewIncomingCall() */


/*
 * MySIPCall::OnHangup()
 * 
 * When hanged up by remote.
 */
int   MySIPCall::OnHangup()
{

  const char *cid;
  simv_info_t *myimvs = NULL;

  cur_call = NULL;
  SIPCall::OnHangup();

  /** Search IM-Voice session by call ID */
  cid = GetCallId();
  debug(DEBUG_MISC,"simvoice","Call closed by remote. [cid=%s]\n", cid);

  myimvs = sip->SearchImvsByCid(cid);
  if(myimvs == NULL) {
    fprintf(stderr, "ERROR: No associated Im-Voice Session!! [cid=%s]\n", cid);
    return 0;
  }

  /* Notify IM user that caller hanged up */
  SIPInstantMessageInfo iminfo; char mesg[MAX_IM_LENGTH];

  sprintf(mesg, "\n*** [SImVoice@%s] ***\n"
                 "  The caller [%s] \n"
                 "  has just hanged up!!\n"
                 "  You can't send a message any more.\n",
                 (const char *)SIPLibrary::HostName4,
                 (const char *)myimvs->voice_user.GetUser());
  iminfo.imbody = mesg;
  iminfo.content_length = strlen(mesg);

  DStringInit(&iminfo.sid);
  DStringAppendD(&iminfo.sid, &myimvs->sid);

  debug(DEBUG_MISC,"simvoice", "Sending IM: to=[%s], cnt=[%s]",
       (const char *)myimvs->im_user, mesg);

  sip->SIPSendMessage(myimvs->im_user, myimvs->server, iminfo);

  /* Let main loop continue to handle */
  cur_imvs    = myimvs;
  simv_status = SIMV_HANGUP_PEND;

  return 0;

} /* OnHangup() */


/**
 * OnEndTalkSpurt()
 * 
 * This call back function 
 * will be called from ASR thread on the end-of-talkspurt event,
 * and this call back should sends out ASR result as SIP-MESSAGE.
 */
int OnEndTalkSpurt(simv_info_t *asr)
{
  SIPInstantMessageInfo iminfo;

  iminfo.imbody = strdup(asr->asr_mesg);
  iminfo.content_length = strlen(asr->asr_mesg);
  DStringInit(&iminfo.sid);
  DStringAppendD(&iminfo.sid, &asr->sid);
  
  debug(DEBUG_MISC,"simvoice","OnEndTalkSpurt: Sending IM: to=[%s], cnt=[%s]\n",
       (const char*)asr->im_user, iminfo.imbody);

  /** Send IM to IM user */
  sip->SIPSendMessage(asr->im_user, asr->server, iminfo);
  free(iminfo.imbody);

  /** Inform voice caller of ASR result */
  char mesg[MAX_IM_LENGTH];

  /*
   * If you want to inform the caller 
   * of the result of ASR using voice
   * use following:
   *
   * sprintf(mesg, "Your speech was sent as %s\n", asr->asr_mesg);
   * SendVoiceMessage(asr, mesg);
   */

  /** If you want to use IM */
  sprintf(mesg, "\n*** [SImVoice@%s] ***\n"
                 " You said: [%s]\n",
                 (const char *)SIPLibrary::HostName4,
                 asr->asr_mesg);
  iminfo.imbody = strdup(mesg);
  iminfo.content_length = strlen(mesg);

  sip->SIPSendMessage(asr->voice_user, asr->server, iminfo);
  free(iminfo.imbody);

  return(0);

} /* OnEndTalkSpurt() */


/* Main routine */
int main(int argc, char *argv[])
{
  int c;
  char arg_line[200];
  unsigned short listening_port = 5060;
  char *asrarg_file = NULL;
  FILE *asrarg_fp = NULL;
  const char *servername = "SImVoice-Columbia-University";

  /* Override SIPLib defaults here ... */
  config.ThreadPoolSize  = 16;
#ifdef USE_TLS
  SIPLibrary::StartSSL = 1;
#endif

  debug(DEBUG_MISC,"simvoice","Simvoice, (c) 2000, Columbia University\n");

  const char *usage = "-a <Sphinx argument file> [options]\n"
    " -v          : Print version and exit\n"
    " -t value    : Set silence detection threshold (Def.7000.0)\n"
    " -d category : verbose mode for the category. (see document)\n"
    "               all,sql,net,sdp,misc,...\n"
    " -o tracefile: put debug trace in this file.\n"
    " -p port     : listening SIP port, default is 5060.\n"
    " -i          : use dotted decimal for self address.\n"
    " -H HostName : IPv6 host name, if different from normal\n";

  /* Parse command line options. */
  while ((c = getopt(argc, argv, "hvt:d:o:a:p:iH:")) != EOF) {
    switch (c) {
      case 'H' :
        SIPLibrary::HostName6 = strdup(optarg);
        break;
      case 'v' :
#ifdef VERSION
        printf("simvoice version %s\n", VERSION);
#else
        printf("simvoice version unknown\n");
#endif
      exit(0);
      break;
      case 't':
        vat_th = atof(optarg);
        break;
      case 'd':
        SIPLibrary::DebugOn();
        DebugAllowModule(optarg, 1);
        break;
      case 'o':
        SIPLibrary::DebugOn(optarg);
        break;
      case 'a':
        asrarg_file = strdup(optarg);
        break;
      case 'p':
        listening_port = atoi(optarg);
        break;
      case 'i' :
        SIPLibrary::oUseNumericAddress = true;
        break;
      case '?':
        /* no break */
      case 'h':
        cerr << "Usage: " << argv[0] << " " << usage << endl;
        exit(1);
    } 
  }
  debug(DEBUG_MISC, "simvoice", "Starting the application\n");

  /** ASR argument file is required */
  if(asrarg_file == NULL){
    cerr << "Usage: " << argv[0] << " " << usage << endl;
    exit(1);
  }

  /** Read argument file for Sphinx2 ASR Engine */
  char *opt, *val, *asr_argv[100];
  int i, len, asr_argc;
  if(!(asrarg_fp = fopen(asrarg_file, "r"))){
    cerr << "Can't find ASR argument file " << asrarg_file << endl; 
    exit(1);
  }
  asr_argc = 0;
  while(1){
    if (fgets(arg_line, sizeof(arg_line), asrarg_fp) == NULL)
      break;
    len = strlen(arg_line);
    if (len >= 1) {
      arg_line[len-1] = 0;
      if (arg_line[0] == '#')
        continue; /* skip comment */
     } else
       continue; 

    len = strlen(arg_line);
    opt = arg_line;
    for(i=0, val=NULL; i<len; i++){
      if(arg_line[i] == ' '){
        arg_line[i] = 0;
        val = &arg_line[i+1];
        break;
      }
    }
    if( val == NULL)
      continue;
    
     asr_argv[asr_argc++] = strdup(opt);
     asr_argv[asr_argc++] = strdup(val);
  }
  if(asrarg_fp != NULL)
    fclose(asrarg_fp);

  /** Initialize ASR Engine */
  if(fbs_init ( asr_argc, asr_argv)){
    fprintf(stderr, "simvoice>> main: failed: fbs_init\n");
    exit( -1);
  }
  debug(DEBUG_MISC,"simvoice","Initialized ASR engine.\n");

  /* Create the listening servers */
  struct sockaddr_storage sin;
  memset(&sin, 0, sizeof(sin));
  sin.ss_family = AF_INET;
  AFISetPort(&sin, listening_port);

  IptelIPAddress listeningAddress(sin);

  sin.ss_family = AF_INET6;
  IptelIPAddress listeningAddress6(sin);
  
  sip = new MySIPUserAgent(listeningAddress, listeningAddress6, false);
  IptelAddress localname = SIPLibrary::HostName4;
  localname.SetUser("simvoice");
  sip->SetLocalName(localname);
 
  /* Initialize debug options */
  SIPLibrary::SetName(servername);
  SIPLibrary::ErrorOn("stderr");

  debug(DEBUG_MISC,"simvoice","Waiting for the first connection..\n");

#ifdef USE_TLS
  sleep(2);             /* the TLS lib needs sometime to start */
#endif

  rtperror the_err; 
  MySIPCall *mycall;

  /** SIMVOICE main loop */
  while (1) {

    if(simv_status == SIMV_ACCEPT_PEND){
      mycall = cur_call;

      /* Accept the Call */
      debug(DEBUG_MISC,"simvoice","Accepting the VOICE call...\n");

      /* Something is wrong */
      if(cur_call == NULL || cur_imvs == NULL){
        debug(DEBUG_MISC,"simvoice","Call or IM is NULL!!");
        simv_status = SIMV_NORMAL;
        continue;
      }

      /* Call accept */
      mycall->Accept();
      debug(DEBUG_MISC,"simvoice","call...Accepted\n");

      /** XXX:
       * Create a ASR thread 
       * which receives voice data from remote,
       * does ASR and sends result as a MESSAGE
       */
      int oRet=0;
      pthread_attr_t attr;

      /* Making attribute for a joinable thread */
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE); 

      oRet = pthread_create(&cur_imvs->asr_thread, &attr, AsrRtpRecvThread, cur_imvs);
      if(oRet)
      {
        cerr << "ERROR: main: pthread_create(ASR)" << endl;
        exit(1);
      }
      pthread_attr_destroy(&attr);

      /* Insert IM-Voice session into the table */
      sip->InsertImvs(cur_imvs);
      simv_status = SIMV_NORMAL;
      cur_call = NULL; cur_imvs = NULL;

      debug(DEBUG_MISC,"simvoice","Waiting for new call...\n");
    } 
    else if(simv_status == SIMV_HANGUP_PEND){

      /* Something is wrong */
      if(cur_imvs == NULL){
        debug(DEBUG_MISC,"simvoice","Current IMVS is NULL!!");
        simv_status = SIMV_NORMAL;
        continue;
      }

      /** Shutdown RTP connection */
      debug(DEBUG_MISC,"simvoice","Terminating ASR Thread...\n");  

      /* This should terminate the thread */
      cur_imvs->oAsrTerminate = true;
      sleep_msec(500);

      the_err = RTPCloseConnection(cur_imvs->rtp_context, "BYE"); 
      the_err = RTPDestroy(cur_imvs->rtp_context);
      
      /* Delete Im-Voice session from the table */
      sip->DeleteImvs(cur_imvs);
      cur_imvs = NULL;
      simv_status = SIMV_NORMAL;

      debug(DEBUG_MISC,"simvoice","Waiting for next connection...\n");

    }
    else if(simv_status == SIMV_REJECT_PEND){
      mycall = cur_call;

      /* Reject the Call */
      debug(DEBUG_MISC,"simvoice","Rejecting the VOICE call...\n");

      /* Call reject */
      mycall->Reject(500, "IM destination unreachable");
      cout << "...Rejected(500)" << endl;

      /** Shutdown RTP session */
      the_err = RTPCloseConnection(cur_imvs->rtp_context, "BYE"); 
      the_err = RTPDestroy(cur_imvs->rtp_context);

      /** Delete Im-Voice session */
      DStringFree(&cur_imvs->cid);
      DStringFree(&cur_imvs->sid);
      delete cur_imvs;
      cur_call = NULL; cur_imvs = NULL;
      simv_status = SIMV_NORMAL;

      debug(DEBUG_MISC,"simvoice","Waiting for next connection...\n");
    }

    /** Sleep for half second just for saving CPU power */
    sleep_msec(500);
    
  }

  /* Never reaches here */
  fbs_end ();
  return 0;
}




