/*
 * tts.c: The thread defined here synthesizes speech
 *        from given text data using IBM ViaVoice TTS engine. 
 *
 * Copyright 2002 by Columbia University; all rights reserved
 */
#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <pthread.h>

#include "time.h"
#include "simv.h"

extern "C"{
unsigned char l16_to_pcmu(int sample);
}


/* Function you must provide for the high-level API */
void RTPSchedule(context cid, rtp_opaque_t opaque, struct timeval *tp){ }


/*
 * sleep_msec()
 * Sleep for specified time in msec
 */ 
void sleep_msec(int32 ms)
{
  timeval tv;
  
  tv.tv_sec = 0;
  tv.tv_usec = ms * 1000; 

  select(0, NULL, NULL, NULL, &tv);
} /* sleep_msec() */


/*
 * TTS_Callback():
 * This callback is called by the ViaVoice library
 * each time when a portion of synthesize is done.
 */
static ECICallbackReturn TTS_Callback(ECIHand eciHand, ECIMessage msg, 
                                      long lparam, void* data)
{
  int samples;
  int i;
  u_char buf[TTS_BUF_SIZE];
  simv_info_t *tts;
  
  /* This means the ViaVoice library is done with a portion */
  if (msg == eciWaveformBuffer)
  {
    debug(DEBUG_MISC, "TTS_Callback", "Called with %ld bytes data.\n", lparam);
     
    /*
     * lparam indicates the number of
     * voice samples synthesized
     */ 
    samples = lparam;
    tts = (simv_info_t *)data;

    /* Check arguments */
    if (tts == NULL || tts->hECI ==  NULL_ECI_HAND) {
      fprintf (stderr, "Error: TTS_CallBack: Buffer/ECI handle is NULL.\n" );
      return eciDataProcessed;
    }
       
    /* Fill in the rest of the buffer with nothing */
    if (TTS_BUF_SIZE > samples)
      memset(&buf[samples], 0x7f, TTS_BUF_SIZE - samples);

    /* l16-->pcmu as SIPC (ratmedia) expects G.711 */
    for (i=0; i<samples && i<TTS_BUF_SIZE; i++)
      buf[i] = l16_to_pcmu((int)tts->tts_buffer[i]);

    /* Send the synthesized voice */
    rtperror the_err;
    int32 tsinc = RTP_SEND_SIZE;
    int32 sent;
    int8 marker = 0;
	
    for(sent=0; sent <= samples-tsinc; sent += tsinc){ 
      the_err = RTPSend(tts->rtp_context, tsinc, marker,
                        PT_MULAW_8, (int8 *)(buf+sent),
                        RTP_SEND_SIZE);

      if ( the_err != RTP_OK )
        fprintf(stderr, "Error: TTS_CallBack: RTPSend: %s\n", RTPStrError(the_err));
      else
        debug(DEBUG_MISC, "TTS_CallBack", "RTPSend: size=%d\n", tsinc);

      /*
       * If the timing block below is removed,
       * ratmedia on the other side will fail
       */
      sleep_msec(20);
    }

    /* If any data left */
    if(sent < samples){
      the_err = RTPSend(tts->rtp_context, tsinc, marker,
                        PT_MULAW_8, (int8 *)(buf+sent),
                        RTP_SEND_SIZE);
      if(the_err != RTP_OK )
        fprintf(stderr, "Error: TTS_CallBack: RTPSend: %s\n", RTPStrError(the_err));
      else
        debug(DEBUG_MISC, "TTS_CallBack", "RTPSend: size=%d\n", tsinc);
    }
  }
  return (eciDataProcessed);

} /* TTS_Callback() */


/*
 * TTS_EnqueueText()
 * Enqueue text data into ViaVoice TTS engine
 * to be synthesized
 */
static void TTS_EnqueueText(simv_info_t *tts, char *text)
{
  int ret;

  /* Check arguments */
  if ( tts == NULL ) {
    fprintf(stderr, "TTS_EnqueueText: argument tts is NULL\n");
    return;
  }
  debug(DEBUG_MISC, "TTS_EnqueueText", "Started with text=%s\n", text);
  
  /* Enqueue text data into TTS engine */
  ret = eciAddText(tts->hECI,text);
  
  /* Begin synthesis */ 
  ret = eciSynthesize(tts->hECI);
  
  /* Synchronized synthesize, blocking function */

  /*
   * This version of RTP library caused error if the read/write from/to
   * the socket happens at the same time, so let's protect by mutex.
   */
  pthread_mutex_lock(&tts->rtp_mutex);
  ret = eciSynchronize(tts->hECI);
  pthread_mutex_unlock(&tts->rtp_mutex);

} /* TTS_EnqueueText() */

/*
 * TtsRtpSendThread()
 *
 * Main TTS thread.
 */
void *TtsRtpSendThread(void *_args)
{

  simv_info_t *tts = (simv_info_t *)_args;
  char *message = tts->tts_mesg;

  int ret;    

  /* Check the arguments */
  if ( tts == NULL ) {
    fprintf(stderr, "TtsRtpSendThread: simv_info_t is NULL\n");
    return NULL;
  }
  debug(DEBUG_MISC, "TtsRtpSendThread", "Started TTS Thread\n");

  /* Create a new ECI instance */
  tts->hECI = eciNew();
  if ( tts->hECI == NULL_ECI_HAND) {
    fprintf(stderr, "TtsRtpSendThread: Couldn't create ECI handle. Check eci.ini\n");
    return NULL;
  }
  
  /* Register the callback function */
  eciRegisterCallback(tts->hECI, TTS_Callback, tts);
  
  /* Sets output buffer of synthesized voice */
  ret = eciSetOutputBuffer(tts->hECI, TTS_BUF_SIZE, tts->tts_buffer);

  /* Set TTS engine to parse annotated text (=1) */
  ret = eciSetParam(tts->hECI, eciInputType, 1);

  /* set TTS engine to sample rate of 0: 8,000Hz, 1: 11,025Hz */
  ret = eciSetParam(tts->hECI, eciSampleRate, 0);
  
  /*
   * Set the voice parameters.
   * Preset voices are as follows:
   *        1:male --> 0:active voice
   *        2:female --> 0:active voice
   *        3:child --> 0:active voice
   *        7:elderly female --> 0:active voice
   *        8:elderly male --> 0:active voice
   */
  ret = eciCopyVoice(tts->hECI, 1, 0);
  
  /*
   * What is this sleep() for?
   * I dared to leave this code, but
   * changed to use select rather than sleep()
   * and much shorter. It was 3 sec before.
   */
  sleep_msec(500);
  
  /* Enqueue text to the engine */
  TTS_EnqueueText(tts, message);

  /** Delete the instance */
  ret = eciSetOutputBuffer(tts->hECI, 0, NULL);
  tts->hECI = eciDelete(tts->hECI);
  
  debug(DEBUG_MISC, "TtsRtpSendThread", "TTS Thread exitting normally.\n");
  pthread_exit((void *) 0);

} /* TtsRtpSendThread() */


